import inspect
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Generic, TypeVar, cast

from jararaca.messagebus.message import INHERITS_MESSAGE_CO, Message, MessageOf
from jararaca.reflect.controller_inspect import (
    ControllerMemberReflect,
    inspect_controller,
)
from jararaca.scheduler.decorators import ScheduledAction, ScheduledActionData

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Any])
DECORATED_T = TypeVar("DECORATED_T", bound=Any)


class MessageHandler(Generic[INHERITS_MESSAGE_CO]):

    MESSAGE_INCOMING_ATTR = "__message_incoming__"

    def __init__(
        self,
        message: type[INHERITS_MESSAGE_CO],
        timeout: int | None = None,
        exception_handler: Callable[[BaseException], None] | None = None,
        nack_on_exception: bool = False,
        auto_ack: bool = True,
        name: str | None = None,
    ) -> None:
        self.message_type = message

        self.timeout = timeout
        self.exception_handler = exception_handler
        self.requeue_on_exception = nack_on_exception
        self.auto_ack = auto_ack
        self.name = name

    def __call__(
        self, func: Callable[[Any, MessageOf[INHERITS_MESSAGE_CO]], Awaitable[None]]
    ) -> Callable[[Any, MessageOf[INHERITS_MESSAGE_CO]], Awaitable[None]]:

        MessageHandler[Any].register(func, self)

        return func

    @staticmethod
    def register(
        func: Callable[[Any, MessageOf[INHERITS_MESSAGE_CO]], Awaitable[None]],
        message_incoming: "MessageHandler[Any]",
    ) -> None:

        setattr(func, MessageHandler.MESSAGE_INCOMING_ATTR, message_incoming)

    @staticmethod
    def get_message_incoming(
        func: Callable[[MessageOf[Any]], Awaitable[Any]],
    ) -> "MessageHandler[Message] | None":
        if not hasattr(func, MessageHandler.MESSAGE_INCOMING_ATTR):
            return None

        return cast(
            MessageHandler[Message], getattr(func, MessageHandler.MESSAGE_INCOMING_ATTR)
        )


@dataclass(frozen=True)
class MessageHandlerData:
    message_type: type[Any]
    spec: MessageHandler[Message]
    instance_callable: Callable[[MessageOf[Any]], Awaitable[None]]
    controller_member: ControllerMemberReflect


@dataclass(frozen=True)
class ScheduleDispatchData:
    timestamp: float


SCHEDULED_ACTION_DATA_SET = set[ScheduledActionData]

MESSAGE_HANDLER_DATA_SET = set[MessageHandlerData]


class MessageBusController:

    MESSAGEBUS_ATTR = "__messagebus__"

    def __init__(self) -> None:
        self.messagebus_factory: (
            Callable[[Any], tuple[MESSAGE_HANDLER_DATA_SET, SCHEDULED_ACTION_DATA_SET]]
            | None
        ) = None

    def get_messagebus_factory(
        self,
    ) -> Callable[
        [DECORATED_T], tuple[MESSAGE_HANDLER_DATA_SET, SCHEDULED_ACTION_DATA_SET]
    ]:
        if self.messagebus_factory is None:
            raise Exception("MessageBus factory is not set")
        return self.messagebus_factory

    def __call__(self, cls_t: type[DECORATED_T]) -> type[DECORATED_T]:

        def messagebus_factory(
            instance: DECORATED_T,
        ) -> tuple[MESSAGE_HANDLER_DATA_SET, SCHEDULED_ACTION_DATA_SET]:
            handlers: MESSAGE_HANDLER_DATA_SET = set()

            schedulers: SCHEDULED_ACTION_DATA_SET = set()

            _, members = inspect_controller(cls_t)

            for name, member in members.items():
                message_handler_decoration = MessageHandler.get_message_incoming(
                    member.member_function
                )
                scheduled_action_decoration = ScheduledAction.get_scheduled_action(
                    member.member_function
                )

                if message_handler_decoration is not None:

                    if not inspect.iscoroutinefunction(member.member_function):
                        raise Exception(
                            "Message incoming handler '%s' from '%s.%s' must be a coroutine function"
                            % (name, cls_t.__module__, cls_t.__qualname__)
                        )

                    handlers.add(
                        MessageHandlerData(
                            message_type=message_handler_decoration.message_type,
                            spec=message_handler_decoration,
                            instance_callable=getattr(instance, name),
                            controller_member=member,
                        )
                    )
                elif scheduled_action_decoration is not None:
                    if not inspect.iscoroutinefunction(member.member_function):
                        raise Exception(
                            "Scheduled action handler '%s' from '%s.%s' must be a coroutine function"
                            % (name, cls_t.__module__, cls_t.__qualname__)
                        )
                    instance_callable = getattr(instance, name)

                    schedulers.add(
                        ScheduledActionData(
                            controller_member=member,
                            spec=scheduled_action_decoration,
                            callable=instance_callable,
                        )
                    )

            return handlers, schedulers

        self.messagebus_factory = messagebus_factory

        MessageBusController.register(cls_t, self)

        return cls_t

    @staticmethod
    def register(func: type[DECORATED_T], messagebus: "MessageBusController") -> None:

        setattr(func, MessageBusController.MESSAGEBUS_ATTR, messagebus)

    @staticmethod
    def get_messagebus(func: type[DECORATED_T]) -> "MessageBusController | None":
        if not hasattr(func, MessageBusController.MESSAGEBUS_ATTR):
            return None

        return cast(
            MessageBusController, getattr(func, MessageBusController.MESSAGEBUS_ATTR)
        )



import asyncio
import logging
import random
from functools import wraps
from typing import Awaitable, Callable, Optional, ParamSpec, TypeVar

logger = logging.getLogger(__name__)

P = ParamSpec("P")
T = TypeVar("T")


class RetryConfig:
    """Configuration for the retry mechanism."""

    def __init__(
        self,
        max_retries: int = 5,
        initial_delay: float = 1.0,
        max_delay: float = 60.0,
        backoff_factor: float = 2.0,
        jitter: bool = True,
    ):
        """
        Initialize retry configuration.

        Args:
            max_retries: Maximum number of retry attempts (default: 5)
            initial_delay: Initial delay in seconds between retries (default: 1.0)
            max_delay: Maximum delay in seconds between retries (default: 60.0)
            backoff_factor: Multiplier for the delay after each retry (default: 2.0)
            jitter: Whether to add randomness to the delay to prevent thundering herd (default: True)
        """
        self.max_retries = max_retries
        self.initial_delay = initial_delay
        self.max_delay = max_delay
        self.backoff_factor = backoff_factor
        self.jitter = jitter


E = TypeVar("E", bound=Exception)


async def retry_with_backoff(
    fn: Callable[[], Awaitable[T]],
    # args: P.args,
    # kwargs: P.kwargs,
    retry_config: Optional[RetryConfig] = None,
    on_retry_callback: Optional[Callable[[int, E, float], None]] = None,
    retry_exceptions: tuple[type[E], ...] = (),
) -> T:
    """
    Execute a function with exponential backoff retry mechanism.

    Args:
        fn: The async function to execute with retry
        *args: Arguments to pass to the function
        retry_config: Configuration for the retry mechanism
        on_retry_callback: Optional callback function called on each retry with retry count, exception, and next delay
        retry_exceptions: Tuple of exception types that should trigger a retry
        **kwargs: Keyword arguments to pass to the function

    Returns:
        The result of the function if successful

    Raises:
        The last exception encountered if all retries fail
    """
    if retry_config is None:
        retry_config = RetryConfig()

    last_exception = None
    delay = retry_config.initial_delay

    for retry_count in range(retry_config.max_retries + 1):
        try:
            return await fn()
        except retry_exceptions as e:
            last_exception = e

            if retry_count >= retry_config.max_retries:
                logger.error(
                    f"Max retries ({retry_config.max_retries}) exceeded: {str(e)}"
                )
                raise

            # Calculate next delay with exponential backoff
            if retry_count > 0:  # Don't increase delay on the first failure
                delay = min(delay * retry_config.backoff_factor, retry_config.max_delay)

            # Apply jitter if configured (±25% randomness)
            if retry_config.jitter:
                jitter_amount = delay * 0.25
                delay = delay + random.uniform(-jitter_amount, jitter_amount)
                # Ensure delay doesn't go negative due to jitter
                delay = max(delay, 0.1)

            logger.warning(
                f"Retry {retry_count+1}/{retry_config.max_retries} after error: {str(e)}. "
                f"Retrying in {delay:.2f}s"
            )

            # Call the optional retry callback if provided
            if on_retry_callback:
                on_retry_callback(retry_count, e, delay)

            await asyncio.sleep(delay)

    # This should never be reached with the current implementation
    if last_exception:
        raise last_exception
    raise RuntimeError("Unexpected error in retry logic")


def with_retry(
    retry_config: Optional[RetryConfig] = None,
    retry_exceptions: tuple[type[Exception], ...] = (Exception,),
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """
    Decorator to wrap an async function with retry logic.

    Args:
        retry_config: Configuration for the retry mechanism
        retry_exceptions: Tuple of exception types that should trigger a retry

    Returns:
        Decorated function with retry mechanism
    """

    def decorator(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @wraps(fn)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return await retry_with_backoff(
                lambda: fn(*args, **kwargs),
                retry_config=retry_config,
                retry_exceptions=retry_exceptions,
            )

        return wrapper

    return decorator



import logging
from contextlib import asynccontextmanager, contextmanager
from typing import AsyncGenerator, Generator, Protocol

from opentelemetry import metrics, trace
from opentelemetry._logs import set_logger_provider
from opentelemetry.baggage.propagation import W3CBaggagePropagator
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as LogExporter,
)
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter as MeterExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as SpanExporter,
)
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from jararaca.microservice import AppTransactionContext, Container, Microservice
from jararaca.observability.decorators import (
    TracingContextProvider,
    TracingContextProviderFactory,
    get_tracing_ctx_provider,
)
from jararaca.observability.interceptor import ObservabilityProvider

tracer: trace.Tracer = trace.get_tracer(__name__)


class OtelTracingContextProvider(TracingContextProvider):

    def __init__(self, app_context: AppTransactionContext) -> None:
        self.app_context = app_context

    @contextmanager
    def __call__(
        self,
        trace_name: str,
        context_attributes: dict[str, str],
    ) -> Generator[None, None, None]:

        with tracer.start_as_current_span(trace_name, attributes=context_attributes):
            yield


class OtelTracingContextProviderFactory(TracingContextProviderFactory):

    def provide_provider(
        self, app_context: AppTransactionContext
    ) -> TracingContextProvider:
        return OtelTracingContextProvider(app_context)

    @asynccontextmanager
    async def root_setup(
        self, app_tx_ctx: AppTransactionContext
    ) -> AsyncGenerator[None, None]:

        title: str = "Unmapped App Context Execution"
        headers = {}
        tx_data = app_tx_ctx.transaction_data
        if tx_data.context_type == "http":

            headers = dict(tx_data.request.headers)
            title = f"HTTP {tx_data.request.method} {tx_data.request.url}"

        elif tx_data.context_type == "message_bus":
            title = f"Message Bus {tx_data.topic}"

        carrier = {
            key: value
            for key, value in headers.items()
            if key.lower().startswith("traceparent")
            or key.lower().startswith("tracestate")
        }

        ctx = TraceContextTextMapPropagator().extract(carrier)

        b2 = {
            key: value
            for key, value in headers.items()
            if key.lower().startswith("baggage")
        }

        ctx2 = W3CBaggagePropagator().extract(b2, context=ctx)

        with tracer.start_as_current_span(name=title, context=ctx2):
            yield


class LoggerHandlerCallback(Protocol):

    def __call__(self, logger_handler: logging.Handler) -> None: ...


class OtelObservabilityProvider(ObservabilityProvider):

    def __init__(
        self,
        app_name: str,
        logs_exporter: LogExporter,
        span_exporter: SpanExporter,
        meter_exporter: MeterExporter,
        logging_handler_callback: LoggerHandlerCallback = lambda _: None,
        meter_export_interval: int = 5000,
    ) -> None:
        self.app_name = app_name
        self.logs_exporter = logs_exporter
        self.span_exporter = span_exporter
        self.meter_exporter = meter_exporter
        self.tracing_provider = OtelTracingContextProviderFactory()
        self.meter_export_interval = meter_export_interval
        self.logging_handler_callback = logging_handler_callback

    @asynccontextmanager
    async def setup(
        self, app: Microservice, container: Container
    ) -> AsyncGenerator[None, None]:
        ### Setup Resource

        resource = Resource(attributes={SERVICE_NAME: self.app_name})

        ### Setup Tracing
        provider = TracerProvider(resource=resource)

        trace.set_tracer_provider(provider)

        span_processor = BatchSpanProcessor(self.span_exporter)
        provider.add_span_processor(span_processor)

        ### Setup Logs
        logger_provider = LoggerProvider(resource=resource)

        set_logger_provider(logger_provider)

        logger_provider.add_log_record_processor(
            BatchLogRecordProcessor(self.logs_exporter)
        )

        logging_handler = LoggingHandler(
            level=logging.DEBUG, logger_provider=logger_provider
        )

        logging_handler.addFilter(lambda _: get_tracing_ctx_provider() is not None)

        self.logging_handler_callback(logging_handler)

        ### Setup Metrics
        metric_reader = PeriodicExportingMetricReader(
            self.meter_exporter, export_interval_millis=self.meter_export_interval
        )
        meter_provider = MeterProvider(metric_readers=[metric_reader])

        metrics.set_meter_provider(meter_provider)

        yield

    @staticmethod
    def from_url(
        app_name: str,
        url: str,
        logging_handler_callback: LoggerHandlerCallback = lambda _: None,
        meter_export_interval: int = 5000,
    ) -> "OtelObservabilityProvider":
        """
        Create an instance of OtelObservabilityProvider with Http Exporters from a given URL
        """

        logs_exporter = LogExporter(endpoint=f"{url}/v1/logs")
        span_exporter = SpanExporter(endpoint=f"{url}/v1/traces")
        metric_exporter = MeterExporter(endpoint=f"{url}/v1/metrics")

        return OtelObservabilityProvider(
            app_name,
            logs_exporter,
            span_exporter,
            metric_exporter,
            logging_handler_callback,
            meter_export_interval,
        )



import asyncio
import contextlib
import logging
import signal
import time
from abc import ABC, abstractmethod
from datetime import UTC, datetime
from typing import Any
from urllib.parse import parse_qs

import aio_pika
import croniter
import urllib3
import urllib3.util
import uvloop
from aio_pika import connect_robust
from aio_pika.abc import AbstractChannel, AbstractRobustConnection
from aio_pika.pool import Pool

from jararaca.broker_backend import MessageBrokerBackend
from jararaca.broker_backend.mapper import get_message_broker_backend_from_url
from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.lifecycle import AppLifecycle
from jararaca.microservice import Microservice
from jararaca.scheduler.decorators import (
    ScheduledAction,
    ScheduledActionData,
    get_type_scheduled_actions,
)
from jararaca.scheduler.types import DelayedMessageData
from jararaca.utils.rabbitmq_utils import RabbitmqUtils

logger = logging.getLogger(__name__)


def _extract_scheduled_actions(
    app: Microservice, container: Container, scheduler_names: set[str] | None = None
) -> list[ScheduledActionData]:
    scheduled_actions: list[ScheduledActionData] = []
    for controllers in app.controllers:

        controller_instance: Any = container.get_by_type(controllers)

        controller_scheduled_actions = get_type_scheduled_actions(controller_instance)

        # Filter scheduled actions by name if scheduler_names is provided
        if scheduler_names is not None:
            filtered_actions = []
            for action in controller_scheduled_actions:
                # Include actions that have a name and it's in the provided set
                if action.spec.name and action.spec.name in scheduler_names:
                    filtered_actions.append(action)
                # Skip actions without names when filtering is active
            controller_scheduled_actions = filtered_actions

        scheduled_actions.extend(controller_scheduled_actions)

    return scheduled_actions


# region Message Broker Dispatcher


class _MessageBrokerDispatcher(ABC):

    @abstractmethod
    async def dispatch_scheduled_action(
        self,
        action_id: str,
        timestamp: int,
    ) -> None:
        """
        Dispatch a message to the message broker.
        This is used to send a message to the message broker
        to trigger the scheduled action.
        """
        raise NotImplementedError("dispatch() is not implemented yet.")

    @abstractmethod
    async def dispatch_delayed_message(
        self,
        delayed_message: DelayedMessageData,
    ) -> None:
        """
        Dispatch a delayed message to the message broker.
        This is used to send a message to the message broker
        to trigger the scheduled action.
        """

        raise NotImplementedError("dispatch_delayed_message() is not implemented yet.")

    @abstractmethod
    async def initialize(self, scheduled_actions: list[ScheduledActionData]) -> None:
        raise NotImplementedError("initialize() is not implemented yet.")

    async def dispose(self) -> None:
        pass


class _RabbitMQBrokerDispatcher(_MessageBrokerDispatcher):

    def __init__(self, url: str) -> None:
        self.url = url

        self.conn_pool: "Pool[AbstractRobustConnection]" = Pool(
            self._create_connection,
            max_size=10,
        )

        self.channel_pool: "Pool[AbstractChannel]" = Pool(
            self._create_channel,
            max_size=10,
        )

        splitted = urllib3.util.parse_url(url)

        assert splitted.scheme in ["amqp", "amqps"], "Invalid URL scheme"

        assert splitted.host, "Invalid URL host"

        assert splitted.query, "Invalid URL query"

        query_params: dict[str, list[str]] = parse_qs(splitted.query)

        assert "exchange" in query_params, "Missing exchange parameter"

        assert query_params["exchange"], "Empty exchange parameter"

        self.exchange = str(query_params["exchange"][0])

    async def _create_connection(self) -> AbstractRobustConnection:
        """
        Create a connection to the RabbitMQ server.
        This is used to send messages to the RabbitMQ server.
        """
        connection = await connect_robust(self.url)
        return connection

    async def _create_channel(self) -> AbstractChannel:
        """
        Create a channel to the RabbitMQ server.
        This is used to send messages to the RabbitMQ server.
        """
        async with self.conn_pool.acquire() as connection:
            channel = await connection.channel()
            return channel

    async def dispatch_scheduled_action(self, action_id: str, timestamp: int) -> None:
        """
        Dispatch a message to the RabbitMQ server.
        This is used to send a message to the RabbitMQ server
        to trigger the scheduled action.
        """

        logger.info(f"Dispatching message to {action_id} at {timestamp}")
        async with self.channel_pool.acquire() as channel:
            exchange = await RabbitmqUtils.get_main_exchange(channel, self.exchange)

            await exchange.publish(
                aio_pika.Message(body=str(timestamp).encode()),
                routing_key=action_id,
            )
            logger.info(f"Dispatched message to {action_id} at {timestamp}")

    async def dispatch_delayed_message(
        self, delayed_message: DelayedMessageData
    ) -> None:
        """
        Dispatch a delayed message to the RabbitMQ server.
        This is used to send a message to the RabbitMQ server
        to trigger the scheduled action.
        """
        async with self.channel_pool.acquire() as channel:

            exchange = await RabbitmqUtils.get_main_exchange(channel, self.exchange)
            await exchange.publish(
                aio_pika.Message(
                    body=delayed_message.payload,
                ),
                routing_key=f"{delayed_message.message_topic}.",
            )

    async def initialize(self, scheduled_actions: list[ScheduledActionData]) -> None:
        """
        Initialize the RabbitMQ server.
        This is used to create the exchange and queues for the scheduled actions.
        """

        async with self.channel_pool.acquire() as channel:
            await RabbitmqUtils.get_main_exchange(channel, self.exchange)

            for sched_act_data in scheduled_actions:
                queue_name = ScheduledAction.get_function_id(sched_act_data.callable)

                # Try to get existing queue
                await RabbitmqUtils.get_scheduled_action_queue(
                    channel=channel,
                    queue_name=queue_name,
                )

    async def dispose(self) -> None:
        await self.channel_pool.close()
        await self.conn_pool.close()


def _get_message_broker_dispatcher_from_url(url: str) -> _MessageBrokerDispatcher:
    """
    Factory function to create a message broker instance from a URL.
    Currently, only RabbitMQ is supported.
    """
    if url.startswith("amqp://") or url.startswith("amqps://"):
        return _RabbitMQBrokerDispatcher(url=url)
    else:
        raise ValueError(f"Unsupported message broker URL: {url}")


# endregion


class BeatWorker:

    def __init__(
        self,
        app: Microservice,
        interval: int,
        broker_url: str,
        backend_url: str,
        scheduled_action_names: set[str] | None = None,
    ) -> None:
        self.app = app

        self.broker: _MessageBrokerDispatcher = _get_message_broker_dispatcher_from_url(
            broker_url
        )
        self.backend: MessageBrokerBackend = get_message_broker_backend_from_url(
            backend_url
        )

        self.interval = interval
        self.scheduler_names = scheduled_action_names
        self.container = Container(self.app)
        self.uow_provider = UnitOfWorkContextProvider(app, self.container)

        self.shutdown_event = asyncio.Event()

        self.lifecycle = AppLifecycle(app, self.container)

    def run(self) -> None:

        def on_shutdown(loop: asyncio.AbstractEventLoop) -> None:
            logger.info("Shutting down - signal received")
            # Schedule the shutdown to run in the event loop
            asyncio.create_task(self._graceful_shutdown())

        with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
            loop = runner.get_loop()
            loop.add_signal_handler(signal.SIGINT, on_shutdown, loop)
            # Add graceful shutdown handler for SIGTERM as well
            loop.add_signal_handler(signal.SIGTERM, on_shutdown, loop)
            runner.run(self.start_scheduler())

    async def start_scheduler(self) -> None:
        """
        Declares the scheduled actions and starts the scheduler.
        This is the main entry point for the scheduler.
        """
        async with self.lifecycle():

            scheduled_actions = _extract_scheduled_actions(
                self.app, self.container, self.scheduler_names
            )

            await self.broker.initialize(scheduled_actions)

            await self.run_scheduled_actions(scheduled_actions)

    async def run_scheduled_actions(
        self, scheduled_actions: list[ScheduledActionData]
    ) -> None:

        while not self.shutdown_event.is_set():
            now = int(time.time())
            for sched_act_data in scheduled_actions:
                func = sched_act_data.callable
                scheduled_action = sched_act_data.spec
                if self.shutdown_event.is_set():
                    break

                async with self.backend.lock():

                    last_dispatch_time: int | None = (
                        await self.backend.get_last_dispatch_time(
                            ScheduledAction.get_function_id(func)
                        )
                    )

                    if last_dispatch_time is not None:
                        cron = croniter.croniter(
                            scheduled_action.cron, last_dispatch_time
                        )
                        next_run: datetime = cron.get_next(datetime).replace(tzinfo=UTC)
                        if next_run > datetime.now(UTC):
                            logger.info(
                                f"Skipping {func.__module__}.{func.__qualname__} until {next_run}"
                            )
                            continue

                    if not scheduled_action.allow_overlap:
                        if (
                            await self.backend.get_in_execution_count(
                                ScheduledAction.get_function_id(func)
                            )
                            > 0
                        ):
                            continue

                    await self.broker.dispatch_scheduled_action(
                        ScheduledAction.get_function_id(func),
                        now,
                    )

                    await self.backend.set_last_dispatch_time(
                        ScheduledAction.get_function_id(func), now
                    )

                    logger.info(
                        f"Scheduled {func.__module__}.{func.__qualname__} at {now}"
                    )

            for (
                delayed_message_data
            ) in await self.backend.dequeue_next_delayed_messages(now):
                await self.broker.dispatch_delayed_message(delayed_message_data)

            with contextlib.suppress(asyncio.TimeoutError):
                await asyncio.wait_for(self.shutdown_event.wait(), self.interval)

            # await self.shutdown_event.wait(self.interval)

        logger.info("Scheduler stopped")

        await self.backend.dispose()
        await self.broker.dispose()

    async def _graceful_shutdown(self) -> None:
        """Handles graceful shutdown process"""
        logger.info("Initiating graceful shutdown sequence")
        self.shutdown_event.set()
        logger.info("Graceful shutdown completed")



import asyncio
import inspect
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Generic, Protocol

from fastapi import APIRouter
from fastapi import Depends as DependsF
from fastapi import WebSocketDisconnect
from fastapi.params import Depends
from fastapi.websockets import WebSocket, WebSocketState
from pydantic import BaseModel

from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.microservice import (
    AppInterceptor,
    AppInterceptorWithLifecycle,
    AppTransactionContext,
    Microservice,
)
from jararaca.presentation.decorators import (
    RestController,
    UseDependency,
    UseMiddleware,
)
from jararaca.presentation.websocket.context import (
    WebSocketConnectionManager,
    WebSocketMessageSender,
    provide_ws_manager,
    provide_ws_message_sender,
)
from jararaca.presentation.websocket.decorators import (
    INHERITS_WS_MESSAGE,
    WebSocketEndpoint,
)

from .base_types import WebSocketMessageBase


class WebSocketMessageWrapper(BaseModel, Generic[INHERITS_WS_MESSAGE]):

    MESSAGE_ID: str
    message: INHERITS_WS_MESSAGE


class BroadcastFunc(Protocol):
    async def __call__(self, message: bytes) -> None: ...


class SendFunc(Protocol):
    async def __call__(self, rooms: list[str], message: bytes) -> None: ...


class WebSocketConnectionBackend(Protocol):

    async def broadcast(self, message: bytes) -> None: ...

    async def send(self, rooms: list[str], message: bytes) -> None: ...

    def configure(
        self, broadcast: BroadcastFunc, send: SendFunc, shutdown_event: asyncio.Event
    ) -> None: ...

    async def shutdown(self) -> None: ...


class WebSocketConnectionManagerImpl(WebSocketConnectionManager):

    def __init__(
        self, backend: WebSocketConnectionBackend, shutdown_event: asyncio.Event
    ) -> None:
        self.rooms: dict[str, set[WebSocket]] = {}
        self.all_websockets: set[WebSocket] = set()
        self.backend = backend
        self.lock = asyncio.Lock()

        self.backend.configure(
            broadcast=self._broadcast_from_backend,
            send=self._send_from_backend,
            shutdown_event=shutdown_event,
        )

    async def broadcast(self, message: bytes) -> None:

        await self.backend.broadcast(message)

    async def _broadcast_from_backend(self, message: bytes) -> None:
        for websocket in self.all_websockets:
            try:
                if websocket.client_state == WebSocketState.CONNECTED:
                    await websocket.send_bytes(message)
            except WebSocketDisconnect:
                async with self.lock:  # TODO: check if this can cause concurrency slowdown issues
                    self.all_websockets.remove(websocket)

    async def send(self, rooms: list[str], message: WebSocketMessageBase) -> None:

        await self.backend.send(
            rooms,
            WebSocketMessageWrapper(MESSAGE_ID=message.MESSAGE_ID, message=message)
            .model_dump_json()
            .encode(),
        )

    async def _send_from_backend(self, rooms: list[str], message: bytes) -> None:
        async with self.lock:
            for room in rooms:
                for websocket in self.rooms.get(room, set()):
                    try:
                        if websocket.client_state == WebSocketState.CONNECTED:
                            await websocket.send_bytes(message)
                    except WebSocketDisconnect:
                        async with self.lock:
                            if websocket in self.rooms[room]:
                                self.rooms[room].remove(websocket)

    async def join(self, rooms: list[str], websocket: WebSocket) -> None:
        for room in rooms:
            self.rooms.setdefault(room, set()).add(websocket)

    async def add_websocket(self, websocket: WebSocket) -> None:
        self.all_websockets.add(websocket)

    async def remove_websocket(self, websocket: WebSocket) -> None:
        self.all_websockets.remove(websocket)
        for room in self.rooms.values():
            room.discard(websocket)

    # async def setup_consumer(self, websocket: WebSocket) -> None: ...


class StagingWebSocketMessageSender(WebSocketMessageSender):

    def __init__(self, connection: WebSocketConnectionManager) -> None:
        self.connection = connection
        self.staged_messages: list[tuple[list[str], WebSocketMessageBase]] = []

    async def send(self, rooms: list[str], message: WebSocketMessageBase) -> None:
        self.staged_messages.append((rooms, message))

    async def flush(self) -> None:
        for rooms, message in self.staged_messages:
            await self.connection.send(rooms, message)
        self.staged_messages.clear()


class WebSocketInterceptor(AppInterceptor, AppInterceptorWithLifecycle):
    """
    WebSocketInterceptor is responsible for managing WebSocket connections and
    intercepting WebSocket requests within the application. It integrates with
    the application's lifecycle and provides a router for WebSocket endpoints.

    Attributes:
        backend (WebSocketConnectionBackend): The backend for managing WebSocket connections.
        shutdown_event (asyncio.Event): Event to signal shutdown.
        connection_manager (WebSocketConnectionManager): Manages WebSocket connections.

    Methods:
        lifecycle(app: Microservice, container: Container) -> AsyncGenerator[None, None]:
            Manages the lifecycle of the WebSocket interceptor.

        intercept(app_context: AppContext) -> AsyncGenerator[None, None]:
            Intercepts WebSocket requests within the application context.

        get_ws_router(app: Microservice, container: Container, uow_provider: UnitOfWorkContextProvider) -> APIRouter:
            Generates an API router for WebSocket endpoints.

    Note:
        This class is deprecated and may be removed in future versions.
    """

    def __init__(self, backend: WebSocketConnectionBackend) -> None:
        self.backend = backend
        self.shutdown_event = asyncio.Event()
        self.connection_manager = WebSocketConnectionManagerImpl(
            backend, self.shutdown_event
        )

    @asynccontextmanager
    async def lifecycle(
        self, app: Microservice, container: Container
    ) -> AsyncGenerator[None, None]:

        yield
        self.shutdown_event.set()

    @asynccontextmanager
    async def intercept(
        self, app_context: AppTransactionContext
    ) -> AsyncGenerator[None, None]:

        staging_ws_messages_sender = StagingWebSocketMessageSender(
            self.connection_manager
        )
        with provide_ws_manager(self.connection_manager), provide_ws_message_sender(
            staging_ws_messages_sender
        ):
            yield
            await staging_ws_messages_sender.flush()

    # def __wrap_with_uow_context_provider(
    #     self, uow: UnitOfWorkContextProvider, func: Callable[..., Any]
    # ) -> Callable[[WebSocket], Awaitable[Any]]:
    #     ctx_manager = uow

    #     @wraps(func)
    #     async def wrapper(ws: WebSocket) -> Any:
    #         async with ctx_manager(WebSocketAppContext(websocket=ws)):
    #             return await func(ws)

    #     return wrapper

    def get_ws_router(
        self,
        app: Microservice,
        container: Container,
        uow_provider: UnitOfWorkContextProvider,
    ) -> APIRouter:
        api_router = APIRouter(
            tags=["WebSocket"],
        )

        for controller_type in app.controllers:

            rest_controller = RestController.get_controller(controller_type)
            controller: Any = container.get_by_type(controller_type)

            members = inspect.getmembers(controller_type, predicate=inspect.isfunction)

            for name, member in members:
                if (ws_endpoint := WebSocketEndpoint.get(member)) is not None:
                    final_path = (
                        rest_controller.path + ws_endpoint.path
                        if rest_controller
                        else ws_endpoint.path
                    )

                    route_dependencies: list[Depends] = []
                    for middlewares_by_hook in UseMiddleware.get_middlewares(
                        getattr(controller_type, name)
                    ):
                        middleware_instance = container.get_by_type(
                            middlewares_by_hook.middleware
                        )
                        route_dependencies.append(
                            Depends(middleware_instance.intercept)
                        )

                    for dependency in UseDependency.get_dependencies(
                        getattr(controller_type, name)
                    ):
                        route_dependencies.append(DependsF(dependency.dependency))

                    api_router.add_api_websocket_route(
                        path=final_path,
                        endpoint=getattr(controller, name),
                        dependencies=route_dependencies,
                        **(ws_endpoint.options or {}),
                    )

        return api_router



from jararaca.broker_backend import MessageBrokerBackend


def get_message_broker_backend_from_url(url: str) -> MessageBrokerBackend:
    """
    Factory function to create a message broker backend instance from a URL.
    Currently, only Redis is supported.
    """
    if (
        url.startswith("redis://")
        or url.startswith("rediss://")
        or url.startswith("redis-socket://")
        or url.startswith("rediss+socket://")
    ):
        from jararaca.broker_backend.redis_broker_backend import (
            RedisMessageBrokerBackend,
        )

        return RedisMessageBrokerBackend(url)
    else:
        raise ValueError(f"Unsupported message broker backend URL: {url}")



import logging
from contextlib import asynccontextmanager
from datetime import datetime, timedelta
from datetime import tzinfo as _TzInfo
from typing import Any, AsyncGenerator

import aio_pika
from aio_pika.abc import AbstractConnection
from pydantic import BaseModel

from jararaca.broker_backend import MessageBrokerBackend
from jararaca.messagebus.interceptors.publisher_interceptor import (
    MessageBusConnectionFactory,
)
from jararaca.messagebus.publisher import IMessage, MessagePublisher
from jararaca.scheduler.types import DelayedMessageData

logger = logging.getLogger(__name__)


class AIOPikaMessagePublisher(MessagePublisher):

    def __init__(
        self,
        channel: aio_pika.abc.AbstractChannel,
        exchange_name: str,
        message_broker_backend: MessageBrokerBackend | None = None,
    ):

        self.channel = channel
        self.exchange_name = exchange_name
        self.message_broker_backend = message_broker_backend
        self.staged_delayed_messages: list[DelayedMessageData] = []
        self.staged_messages: list[IMessage] = []

    async def publish(self, message: IMessage, topic: str) -> None:
        self.staged_messages.append(message)

    async def _publish(self, message: IMessage, topic: str) -> None:
        exchange = await self.channel.get_exchange(self.exchange_name, ensure=False)
        if not exchange:
            logging.warning(f"Exchange {self.exchange_name} not found")
            return
        routing_key = f"{topic}."
        await exchange.publish(
            aio_pika.Message(body=message.model_dump_json().encode()),
            routing_key=routing_key,
        )

    async def delay(self, message: IMessage, seconds: int) -> None:
        if not self.message_broker_backend:
            raise NotImplementedError(
                "Delay is not implemented for AIOPikaMessagePublisher"
            )
        self.staged_delayed_messages.append(
            DelayedMessageData(
                message_topic=message.MESSAGE_TOPIC,
                payload=message.model_dump_json().encode(),
                dispatch_time=int(
                    (datetime.now(tz=None) + timedelta(seconds=seconds)).timestamp()
                ),
            )
        )

    async def schedule(self, message: IMessage, when: datetime, tz: _TzInfo) -> None:
        if not self.message_broker_backend:
            raise NotImplementedError(
                "Schedule is not implemented for AIOPikaMessagePublisher"
            )
        self.staged_delayed_messages.append(
            DelayedMessageData(
                message_topic=message.MESSAGE_TOPIC,
                payload=message.model_dump_json().encode(),
                dispatch_time=int(when.timestamp()),
            )
        )

    async def flush(self) -> None:
        for message in self.staged_messages:
            logger.debug(
                f"Publishing message {message.MESSAGE_TOPIC} with payload: {message.model_dump_json()}"
            )
            await self._publish(message, message.MESSAGE_TOPIC)

        if len(self.staged_delayed_messages) > 0:
            if not self.message_broker_backend:
                raise NotImplementedError(
                    "MessageBrokerBackend is required to publish delayed messages"
                )

            for delayed_message in self.staged_delayed_messages:
                logger.debug(
                    f"Scheduling delayed message {delayed_message.message_topic} with payload: {delayed_message.payload.decode()}"
                )
                await self.message_broker_backend.enqueue_delayed_message(
                    delayed_message
                )


class GenericPoolConfig(BaseModel):
    max_size: int


class AIOPikaConnectionFactory(MessageBusConnectionFactory):

    def __init__(
        self,
        url: str,
        exchange: str,
        connection_pool_config: GenericPoolConfig | None = None,
        channel_pool_config: GenericPoolConfig | None = None,
        message_broker_backend: MessageBrokerBackend | None = None,
    ):
        self.url = url
        self.exchange = exchange
        self.message_broker_backend = message_broker_backend
        self.connection_pool: aio_pika.pool.Pool[AbstractConnection] | None = None
        self.channel_pool: aio_pika.pool.Pool[aio_pika.abc.AbstractChannel] | None = (
            None
        )

        if connection_pool_config:

            async def get_connection() -> AbstractConnection:
                return await aio_pika.connect_robust(self.url)

            self.connection_pool = aio_pika.pool.Pool[AbstractConnection](
                get_connection,
                max_size=connection_pool_config.max_size,
            )

        if channel_pool_config:

            async def get_channel() -> aio_pika.abc.AbstractChannel:
                async with self.acquire_connection() as connection:
                    return await connection.channel(publisher_confirms=False)

            self.channel_pool = aio_pika.pool.Pool[aio_pika.abc.AbstractChannel](
                get_channel, max_size=channel_pool_config.max_size
            )

    @asynccontextmanager
    async def acquire_connection(self) -> AsyncGenerator[AbstractConnection, Any]:
        if not self.connection_pool:
            async with await aio_pika.connect_robust(self.url) as connection:
                yield connection
        else:

            async with self.connection_pool.acquire() as connection:
                yield connection

    @asynccontextmanager
    async def acquire_channel(
        self,
    ) -> AsyncGenerator[aio_pika.abc.AbstractChannel, Any]:
        if not self.channel_pool:
            async with self.acquire_connection() as connection:
                yield await connection.channel(publisher_confirms=False)
        else:
            async with self.channel_pool.acquire() as channel:
                yield channel

    @asynccontextmanager
    async def provide_connection(self) -> AsyncGenerator[AIOPikaMessagePublisher, Any]:

        async with self.acquire_channel() as channel:
            tx = channel.transaction()

            await tx.select()

            try:
                yield AIOPikaMessagePublisher(
                    channel,
                    exchange_name=self.exchange,
                    message_broker_backend=self.message_broker_backend,
                )
                await tx.commit()
            except Exception as e:
                await tx.rollback()
                raise e



from contextlib import asynccontextmanager
from typing import AsyncContextManager, AsyncGenerator, Protocol

from jararaca.broker_backend import MessageBrokerBackend
from jararaca.messagebus.publisher import MessagePublisher, provide_message_publisher
from jararaca.microservice import AppInterceptor, AppTransactionContext


class MessageBusConnectionFactory(Protocol):

    def provide_connection(self) -> AsyncContextManager[MessagePublisher]: ...


class MessageBusPublisherInterceptor(AppInterceptor):

    def __init__(
        self,
        connection_factory: MessageBusConnectionFactory,
        connection_name: str = "default",
        message_scheduler: MessageBrokerBackend | None = None,
    ):
        self.connection_factory = connection_factory
        self.connection_name = connection_name
        self.message_scheduler = message_scheduler

    @asynccontextmanager
    async def intercept(
        self, app_context: AppTransactionContext
    ) -> AsyncGenerator[None, None]:
        if app_context.transaction_data.context_type == "websocket":
            yield
            return

        async with self.connection_factory.provide_connection() as connection:
            with provide_message_publisher(self.connection_name, connection):
                yield

                await connection.flush()



from jararaca.presentation.websocket.base_types import WebSocketMessageBase
from jararaca.presentation.websocket.context import use_ws_message_sender


class WebSocketMessage(WebSocketMessageBase):

    async def send(self, *rooms: str) -> None:
        await use_ws_message_sender().send(list(rooms), self)



from contextlib import asynccontextmanager, contextmanager, suppress
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Any, AsyncGenerator, Generator

from sqlalchemy.ext.asyncio import (
    AsyncSession,
    AsyncSessionTransaction,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.ext.asyncio.engine import AsyncEngine

from jararaca.microservice import AppInterceptor, AppTransactionContext
from jararaca.reflect.metadata import SetMetadata, get_metadata_value

DEFAULT_CONNECTION_NAME = "default"

ctx_default_connection_name: ContextVar[str] = ContextVar(
    "ctx_default_connection_name", default=DEFAULT_CONNECTION_NAME
)


def ensure_name(name: str | None) -> str:
    return ctx_default_connection_name.get()


@dataclass
class PersistenceCtx:
    session: AsyncSession
    tx: AsyncSessionTransaction


ctx_session_map = ContextVar[dict[str, PersistenceCtx]]("ctx_session_map", default={})


@contextmanager
def providing_session(
    session: AsyncSession,
    tx: AsyncSessionTransaction,
    connection_name: str | None = None,
) -> Generator[None, Any, None]:
    """
    Context manager to provide a session and transaction for a specific connection name.
    If no connection name is provided, it uses the default connection name from the context variable.
    """
    connection_name = ensure_name(connection_name)
    current_map = ctx_session_map.get({})

    token = ctx_session_map.set(
        {**current_map, connection_name: PersistenceCtx(session, tx)}
    )

    try:
        yield
    finally:
        with suppress(ValueError):
            ctx_session_map.reset(token)


provide_session = providing_session
"""
Alias for `providing_session` to maintain backward compatibility.
"""


@asynccontextmanager
async def providing_new_session(
    connection_name: str | None = None,
) -> AsyncGenerator[AsyncSession, None]:

    current_session = use_session(connection_name)

    async with AsyncSession(
        current_session.bind,
    ) as new_session, new_session.begin() as new_tx:
        with providing_session(new_session, new_tx, connection_name):
            yield new_session


def use_session(connection_name: str | None = None) -> AsyncSession:
    connection_name = ensure_name(connection_name)
    current_map = ctx_session_map.get({})
    if connection_name not in current_map:
        raise ValueError(
            f'Session not found for connection "{connection_name}" in context. Check if your interceptor is correctly set up.'
        )

    return current_map[connection_name].session


@contextmanager
def providing_transaction(
    tx: AsyncSessionTransaction,
    connection_name: str | None = None,
) -> Generator[None, Any, None]:
    connection_name = ensure_name(connection_name)

    current_map = ctx_session_map.get({})

    if connection_name not in current_map:
        raise ValueError(f"No session found for connection {connection_name}")

    with providing_session(current_map[connection_name].session, tx, connection_name):
        yield


def use_transaction(connection_name: str | None = None) -> AsyncSessionTransaction:
    current_map = ctx_session_map.get({})
    if connection_name not in current_map:
        raise ValueError(f"Transaction not found for connection {connection_name}")

    return current_map[connection_name].tx


class AIOSQAConfig:
    url: str | AsyncEngine
    connection_name: str
    inject_default: bool

    def __init__(
        self,
        url: str | AsyncEngine,
        connection_name: str = DEFAULT_CONNECTION_NAME,
        inject_default: bool = True,
    ):
        self.url = url
        self.connection_name = connection_name
        self.inject_default = inject_default


INJECT_CONNECTION_METADATA = "inject_connection_metadata_{connection_name}"


def set_inject_connection(
    inject: bool, connection_name: str = DEFAULT_CONNECTION_NAME
) -> SetMetadata:
    """
    Set whether to inject the connection metadata for the given connection name.
    This is useful when you want to control whether the connection metadata
    should be injected into the context or not.
    """

    return SetMetadata(
        INJECT_CONNECTION_METADATA.format(connection_name=connection_name), inject
    )


def uses_connection(
    connection_name: str = DEFAULT_CONNECTION_NAME,
) -> SetMetadata:
    """
    Use connection metadata for the given connection name.
    This is useful when you want to inject the connection metadata into the context,
    for example, when you are using a specific connection for a specific operation.
    """
    return SetMetadata(
        INJECT_CONNECTION_METADATA.format(connection_name=connection_name), True
    )


def dnt_uses_connection(
    connection_name: str = DEFAULT_CONNECTION_NAME,
) -> SetMetadata:
    """
    Do not use connection metadata for the given connection name.
    This is useful when you want to ensure that the connection metadata is not injected
    into the context, for example, when you are using a different connection for a specific operation.
    """
    return SetMetadata(
        INJECT_CONNECTION_METADATA.format(connection_name=connection_name), False
    )


class AIOSqlAlchemySessionInterceptor(AppInterceptor):

    def __init__(self, config: AIOSQAConfig):
        self.config = config
        self.engine = (
            create_async_engine(self.config.url)
            if isinstance(self.config.url, str)
            else self.config.url
        )

        self.sessionmaker = async_sessionmaker(self.engine)

    @asynccontextmanager
    async def intercept(
        self, app_context: AppTransactionContext
    ) -> AsyncGenerator[None, None]:

        uses_connection_metadata = get_metadata_value(
            INJECT_CONNECTION_METADATA.format(
                connection_name=self.config.connection_name
            ),
            self.config.inject_default,
        )

        if not uses_connection_metadata:
            yield
            return

        async with self.sessionmaker() as session, session.begin() as tx:
            token = ctx_default_connection_name.set(self.config.connection_name)
            with providing_session(session, tx, self.config.connection_name):
                try:
                    yield
                    if tx.is_active:
                        await tx.commit()
                except Exception as e:
                    await tx.rollback()
                    raise e
                finally:
                    with suppress(ValueError):
                        ctx_default_connection_name.reset(token)



from typing import Any, Callable, TypeVar

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Any])


class QueryEndpoint:
    """
    Decorator to mark a endpoint function as a query endpoint for Typescript generation.
    """

    METADATA_KEY = "__jararaca_query_endpoint__"

    def __init__(self) -> None: ...

    def __call__(self, func: DECORATED_FUNC) -> DECORATED_FUNC:
        """
        Decorate the function to mark it as a query endpoint.
        """
        setattr(func, self.METADATA_KEY, True)
        return func

    @staticmethod
    def is_query(func: Any) -> bool:
        """
        Check if the function is marked as a query endpoint.
        """
        return getattr(func, QueryEndpoint.METADATA_KEY, False)


class MutationEndpoint:
    """
    Decorator to mark a endpoint function as a mutation endpoint for Typescript generation.
    """

    METADATA_KEY = "__jararaca_mutation_endpoint__"

    def __init__(self) -> None: ...

    def __call__(self, func: DECORATED_FUNC) -> DECORATED_FUNC:
        """
        Decorate the function to mark it as a mutation endpoint.
        """
        setattr(func, self.METADATA_KEY, True)
        return func

    @staticmethod
    def is_mutation(func: Any) -> bool:
        """
        Check if the function is marked as a mutation endpoint.
        """
        return getattr(func, MutationEndpoint.METADATA_KEY, False)



import time
from urllib.parse import urljoin

import httpx

from jararaca.rpc.http.decorators import (
    HttpRPCAsyncBackend,
    HttpRPCRequest,
    HttpRPCResponse,
    RPCRequestNetworkError,
    TimeoutException,
)


class HTTPXHttpRPCAsyncBackend(HttpRPCAsyncBackend):

    def __init__(self, prefix_url: str = "", default_timeout: float = 30.0):
        self.prefix_url = prefix_url
        self.default_timeout = default_timeout

    async def request(
        self,
        request: HttpRPCRequest,
    ) -> HttpRPCResponse:

        start_time = time.time()

        # Prepare timeout
        timeout = (
            request.timeout if request.timeout is not None else self.default_timeout
        )

        # Prepare request kwargs
        request_kwargs = {
            "method": request.method,
            "url": urljoin(self.prefix_url, request.url),
            "headers": request.headers,
            "params": request.query_params,
            "timeout": timeout,
        }

        # Handle different content types
        if request.form_data and request.files:
            # Multipart form data with files
            request_kwargs["data"] = request.form_data
            request_kwargs["files"] = request.files
        elif request.form_data:
            # Form data only
            request_kwargs["data"] = request.form_data
        elif request.body:
            # Raw body content
            request_kwargs["content"] = request.body

        async with httpx.AsyncClient() as client:
            try:
                response = await client.request(**request_kwargs)  # type: ignore[arg-type]

                elapsed_time = time.time() - start_time

                return HttpRPCResponse(
                    status_code=response.status_code,
                    data=response.content,
                    headers=dict(response.headers),
                    elapsed_time=elapsed_time,
                )
            except httpx.TimeoutException as err:
                raise TimeoutException(f"Request timed out: {err}") from err
            except httpx.NetworkError as err:
                raise RPCRequestNetworkError(
                    request=request, backend_request=err.request
                ) from err
